"""Return the minkowsky distance beetween the given dictionaries."""
from .nth_variation import nth_variation
from typing import Dict


def minkowsky(a: Dict, b: Dict, p: int = 2) -> float:
    """Return the minkowsky distance beetween the given dictionaries.

    Parameters
    ----------------------------
    a: Dict,
        First dictionary to consider.
    b: Dict,
        Second dictionary to consider.
    P: int,
        The order of the norm of the difference.

    Raises
    ----------------------------
    ValueError:
        When the order of the norm of the difference is zero.

    Returns
    ----------------------------
    Return the minkowsky distance beetween the given dictionaries.
    """
    if p == 0:
        raise ValueError("Parameter p must be non zero.")
    return nth_variation(a, b, p)**(1 / p)



"""Return the euclidean distance beetween the given dictionaries."""
from .minkowsky import minkowsky
from typing import Dict


def euclidean(a: Dict, b: Dict)->float:
    """Return the euclidean distance beetween the given dictionaries.

    Parameters
    ----------------------------
    a: Dict,
        First dictionary to consider.
    b: Dict,
        Second dictionary to consider.

    Returns
    ----------------------------
    Return the euclidean distance beetween the given dictionaries.
    """
    return minkowsky(a, b, 2)



"""Return the mean absolute error beetween the given dictionaries."""
from .distances_utils import mean
from .total_variation import total_variation
from typing import Dict


def mae(a:Dict, b:Dict)->float:
    """Return the mean absolute error beetween the given dictionaries.

    Parameters
    ----------------------------
    a: Dict,
        First dictionary to consider.
    b: Dict,
        Second dictionary to consider.

    Returns
    ----------------------------
    Return the mean absolute error beetween the given dictionaries.
    """
    return mean(a, b, total_variation)



"""Return the Total Variation distance."""
from .nth_variation import nth_variation
from typing import Dict


def total_variation(a: Dict, b: Dict, overlap: bool = False) -> float:
    """Return the Total Variation distance."""
    return nth_variation(a, b, 1, overlap)



"""Return the normalized chi square distance beetween a and b."""
from typing import Dict

def normal_chi_square(a: Dict, b: Dict)->float:
    """Return the normalized chi square distance beetween a and b."""

    a_factor = 1.0/sum(a.values())
    b_factor = 1.0/sum(b.values())
    total = 0
    for k, aval in a.items():
        aval = aval*a_factor
        try:
            bval = b[k]*b_factor
            total += (aval - bval) * (aval - bval) / (aval + bval)
        except KeyError:
            total += aval

    for k in b:
        try:
            a[k]
        except KeyError:
            total += b[k]*b_factor

    return total



"""Return the nth power distance beetween the given dictionaries."""
from typing import Dict


def _even_nth(a_val: float, b_val: float) -> float:
    return a_val - b_val


def _odd_nth(a_val: float, b_val: float) -> float:
    return abs(a_val - b_val)


def nth_variation(a: Dict, b: Dict, exp: float = 2, overlap: bool = False) -> float:
    """Return the nth power distance beetween the given dictionaries.

    Parameters
    ----------------------------
    a: Dict,
        First dictionary to consider.
    b: Dict,
        Second dictionary to consider.
    exp: float,
        The exponent for the nth power distance.
    overlap: bool,
        Whetever to return or not the overlap number.

    Returns
    ----------------------------
    Return the nth power distance distance beetween the given dictionaries.
    """
    total = 0
    n = 0
    bget = b.__getitem__
    aget = a.__getitem__

    if exp % 2 == 0:
        nth = _even_nth
    else:
        nth = _odd_nth

    for k, a_val in a.items():
        try:
            total += nth(a_val, bget(k))**exp
        except KeyError:
            total += a_val**exp
        n += 1
    for k, b_val in b.items():
        try:
            aget(k)
        except KeyError:
            total += b_val**exp
            n += 1
    result = total
    if overlap:
        return result, n
    return result



"""Return the chi square distance beetween a and b."""
from typing import Dict

def chi_square(a: Dict, b: Dict)->float:
    """Return the chi square distance beetween a and b."""

    total = 0
    for k, aval in a.items():
        try:
            bval = b[k]
            total += (aval - bval) * (aval - bval) / (aval + bval)
        except KeyError:
            total += aval

    for k in b:
        try:
            a[k]
        except KeyError:
            total += b[k]

    return total



"""Import all the functions offered by package dictances."""
from .bhattacharyya import bhattacharyya, bhattacharyya_coefficient
from .canberra import canberra
from .chebyshev import chebyshev
from .chi_square import chi_square
from .cosine import cosine
from .euclidean import euclidean
from .hamming import hamming
from .intersection_squared_hellinger import intersection_squared_hellinger
from .intersection_squared_variation import intersection_squared_variation
from .intersection_total_variation import intersection_total_variation
from .intersection_nth_variation import intersection_nth_variation
from .jensen_shannon import jensen_shannon
from .kullback_leibler import kullback_leibler
from .mae import mae
from .minkowsky import minkowsky
from .mse import mse
from .normal_chi_square import normal_chi_square
from .nth_variation import nth_variation
from .pearson import pearson
from .squared_variation import squared_variation
from .total_variation import total_variation

manhattan = cityblock = total_variation

__all__ = ['bhattacharyya', 'bhattacharyya_coefficient', 'canberra',
           'chebyshev', 'chi_square', 'cosine', 'euclidean',
           'hamming', 'jensen_shannon', 'kullback_leibler',
           'mae', 'manhattan', 'cityblock', 'minkowsky', 'mse', 'normal_chi_square',
           'nth_variation', 'pearson', 'squared_variation', 'total_variation', 'intersection_squared_variation',
           'intersection_squared_hellinger', 'intersection_nth_variation']



"""Return the squared distance beetween a and b."""
from .nth_variation import nth_variation
from typing import Dict

def squared_variation(a: Dict, b: Dict, overlap: bool=False)->float:
    """Return the squared distance beetween a and b."""
    return nth_variation(a, b, 2, overlap)



"""Return the Jensen Shannon Divergence beetween the given dictionaries."""
from math import log
from typing import Dict
from .distances_utils import sort


def jensen_shannon(a: Dict, b: Dict)->float:
    """Return the Jensen Shannon divergence beetween the given dictionaries.

    Parameters
    ----------------------------
    a: Dict,
        First dictionary to consider.
    b: Dict,
        Second dictionary to consider.

    Returns
    ----------------------------
    Return the Jensen Shannon divergence beetween the given dictionaries.
    """
    total = 0
    delta = 0
    big, small = sort(a, b)

    big_get = big.__getitem__

    for key, small_value in small.items():
        try:
            big_value = big_get(key)
            if big_value:
                denominator = (big_value + small_value) / 2
                total += small_value * log(small_value / denominator) + \
                    big_value * log(big_value / denominator)
                delta += big_value + small_value
        except KeyError:
            pass

    total += (2 - delta) * log(2)
    return total / 2



"""Determine the Intersection Squared Hellinger distance."""
from math import sqrt
from typing import Dict
from .distances_utils import sort


def intersection_squared_hellinger(a: Dict, b: Dict) -> float:
    """Determine the Intersection Squared Hellinger distance."""
    total = 0
    big, small = sort(a, b)
    big_get = big.__getitem__
    for key, small_value in small.items():
        try:
            total += (sqrt(small_value) - sqrt(big_get(key)))**2
        except KeyError:
            pass
    return total



"""Return the canberra distance beetween the given dictionaries."""
from typing import Dict


def canberra(a: Dict, b: Dict) -> float:
    """Return the canberra distance beetween the given dictionaries.

    Parameters
    ----------------------------
    a: Dict,
        First dictionary to consider.
    b: Dict,
        Second dictionary to consider.

    Returns
    ----------------------------
    Return the canberra distance beetween the given dictionaries.
    """
    total = 0
    bget = b.__getitem__
    aget = a.__getitem__
    for k, aval in a.items():
        try:
            bval = bget(k)
            total += abs(aval - bval) / (aval + bval)
        except KeyError:
            total += 1
    for k in b:
        try:
            aget(k)
        except KeyError:
            total += 1
    return total



"""Determine the Kullback Leibler divergence beetween the given dictionaries."""
from math import log, inf
from typing import Dict
from .distances_utils import sort


def kullback_leibler(a: Dict, b: Dict) -> float:
    """Determine the Kullback Leibler divergence beetween the given dictionaries.

    Parameters
    ----------------------------
    a: Dict,
        First dictionary to consider.
    b: Dict,
        Second dictionary to consider.

    Returns
    ----------------------------
    Return the Kullback Leibler divergence beetween the given dictionaries.
    """
    total = 0
    overlap = 0
    big, small = sort(a, b)
    big_get = big.__getitem__
    for key, small_value in small.items():
        try:
            total += small_value * log(small_value / big_get(key))
            overlap += 1
        except KeyError:
            pass

    if overlap == 0 and (len(a) != 0 or len(b) != 0):
        return inf
    return total



"""Return the Hamming distance beetween the given dictionaries."""
from typing import Dict

def hamming(a: Dict, b: Dict)->float:
    """Return the Hamming distance beetween the given dictionaries.

    Parameters
    ----------------------------
    a: Dict,
        First dictionary to consider.
    b: Dict,
        Second dictionary to consider.

    Returns
    ----------------------------
    Return the Hamming distance beetween the given dictionaries.
    """
    common = 0
    bget = b.__getitem__
    for k in a:
        try:
            bget(k)
            common += 1
        except KeyError:
            pass
    return len(a) + len(b) - common * 2



"""Return the cosine distance beetween the given dictionaries."""
from math import sqrt
from typing import Dict

def cosine(a: Dict, b: Dict)->float:
    """Return the cosine distance beetween the given dictionaries.

    Parameters
    ----------------------------
    a: Dict,
        First dictionary to consider.
    b: Dict,
        Second dictionary to consider.

    Returns
    ----------------------------
    Return the cosine distance beetween the given dictionaries.
    """
    prod_ab = 0
    sqr_a = 0
    sqr_b = 0
    bget = b.__getitem__
    for k, a_val in a.items():
        sqr_a += a_val**2
        try:
            prod_ab += bget(k) * a_val
        except KeyError:
            pass

    for k, b_val in b.items():
        sqr_b += b_val**2

    return 1 - prod_ab / (sqrt(sqr_a) * sqrt(sqr_b))



"""Determine the bhattacharyya distance beetween the given dictionaries."""
import numpy as np
from math import sqrt
from typing import Dict
from .distances_utils import sort


def bhattacharyya_coefficient(a: Dict, b: Dict) -> float:
    """Determine the Bhattacharyya coefficient beetween the given dictionaries.

    Parameters
    ----------------------------
    a: Dict,
        First dictionary to consider.
    b: Dict,
        Second dictionary to consider.

    Returns
    ----------------------------
    Return the Bhattacharyya coefficient beetween the given dictionaries.
    """
    total = 0
    big, small = sort(a, b)
    big_get = big.__getitem__
    for k, small_value in small.items():
        try:
            total += sqrt(big_get(k) * small_value)
        except KeyError:
            pass
    return total


def bhattacharyya(a: Dict, b: Dict) -> float:
    """Determine the Bhattacharyya distance beetween the given dictionaries.

    Parameters
    ----------------------------
    a: Dict,
        First dictionary to consider.
    b: Dict,
        Second dictionary to consider.

    Returns
    ----------------------------
    Return the Bhattacharyya distance beetween the given dictionaries.
    """
    distance = -np.log(bhattacharyya_coefficient(a, b))
    if np.isinf(distance):
        return 0
    return distance


"""Generic utils used through all the package."""
from typing import Dict

def sort(a: Dict, b: Dict):
    """Return the bigger dict, then the smaller one."""
    if len(a) > len(b):
        return a, b
    return b, a


def mean(a, b, distance):
    """Return the mean of the given distance."""
    result, n = distance(a, b, overlap=True)
    return result / n


"""Return the nth power distance beetween a and b."""
from math import sqrt
from typing import Dict

def pearson(a: Dict, b: Dict)->float:
    """Return the nth power distance beetween a and b."""
    bget = b.get

    ab = 0
    n_mul = 0
    a_sum = 0
    b_sum = 0
    a_sum2 = 0
    b_sum2 = 0

    for k, a_val in a.items():
        b_val = bget(k)
        a_sum += a_val
        a_sum2 += a_val**2
        if b_val:
            n_mul += 1
            ab += a_val * b_val

    for k, b_val in b.items():
        b_sum += b_val
        b_sum2 += b_val**2

    len_a = len(a)
    len_b = len(b)

    a_mean = a_sum / len_a
    b_mean = b_sum / len_b

    return 1 - (ab - n_mul * a_mean * b_mean) / (sqrt(a_sum2 - len_a * a_mean**2) * sqrt(b_sum2 - len_b * b_mean**2))



"""Return the nth power distance beetween the insersection of a and b."""
from typing import Dict

def _even_nth(a_val: float, b_val: float)->float:
    return a_val - b_val


def _odd_nth(a_val: float, b_val: float)->float:
    return abs(a_val - b_val)


def intersection_nth_variation(a: Dict, b: Dict, exp: float=2, overlap: bool=False)->float:
    """Return the nth power distance beetween the insersection of a and b."""
    total = 0
    n = 0
    bget = b.__getitem__

    if exp % 2 == 0:
        nth = _even_nth
    else:
        nth = _odd_nth

    for k, a_val in a.items():
        try:
            total += nth(a_val, bget(k))**exp
            n += 1
        except KeyError:
            pass
    result = total
    if overlap:
        return result, n
    return result



"""Return the squared distance beetween the intersection of a and b."""
from .intersection_nth_variation import intersection_nth_variation
from typing import Dict

def intersection_squared_variation(a: Dict, b: Dict, overlap: bool=False)->float:
    """Return the squared distance beetween the intersection of a and b."""
    return intersection_nth_variation(a, b, 2, overlap)



"""Return the mean total squared variation."""
from .distances_utils import mean
from .squared_variation import squared_variation
from typing import Dict


def mse(a: Dict, b: Dict) -> float:
    """Return the mean squared error beetween the given dictionaries.

    Parameters
    ----------------------------
    a: Dict,
        First dictionary to consider.
    b: Dict,
        Second dictionary to consider.

    Returns
    ----------------------------
    Return the mean squared error beetween the given dictionaries.
    """
    return mean(a, b, squared_variation)



"""Determine the Chebyshev distance beetween the given dictionaries."""
from typing import Dict


def chebyshev(a: Dict, b: Dict) -> float:
    """Determine the Chebyshev distance beetween the given dictionaries.

    Parameters
    ----------------------------
    a: Dict,
        First dictionary to consider.
    b: Dict,
        Second dictionary to consider.

    Returns
    ----------------------------
    Return the Chebyshev distance beetween the given dictionaries.
    """
    result = 0
    bget = b.__getitem__
    aget = a.__getitem__
    for key, a_val in a.items():
        try:
            result = max(result, abs(a_val - bget(key)))
        except KeyError:
            result = max(result, a_val)
    for key, b_val in b.items():
        try:
            aget(key)
        except KeyError:
            result = max(result, b_val)
    return result
